using MeshVR;
using MVR.FileManagementSecure;
using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Policy;
using UnityEngine;

public class FaceBlend : MVRScript
{
    private const string ASSET_BUNDLE_LOCATION = "Custom/Assets/northernshikima/northernshikima_faceblend.assetbundle";
    private const string DEFAULT_MASK_PATH = "Custom/Atom/Person/Textures/FaceBlend Masks/";
    private const string DEFAULT_MASK_TEXTURE = "Face Blend Mask.png";
    private Material[] materials;

    private Texture originalFaceTexture;
    private Texture secondTexture;
    private Texture maskTexture;

    private static int[] faceMaterialSlots = new int[] { 2, 5, 11 };

    // Static so they're only loaded once
    private Shader unlitBlendShader;
    private bool assetsLoaded = false;

    private JSONStorableFloat hue = new JSONStorableFloat("Hue", 0.5f, 0.0f, 1.0f);
    private JSONStorableFloat saturation = new JSONStorableFloat("Saturation", 0.5f, 0.0f, 1.0f);
    private JSONStorableFloat value = new JSONStorableFloat("Value", 0.5f, 0.0f, 1.0f);

    private JSONStorableUrl secondTextureURL = new JSONStorableUrl("FaceBlend Second Texture", string.Empty);
    private JSONStorableUrl maskTextureURL =  new JSONStorableUrl("FaceBlend Mask Texture", string.Empty);

    private RenderTexture currentRenderTexture;

    public override void Init()
    {
        try
        {
            if (containingAtom == null || containingAtom.type != "Person")
            {
                SuperController.LogError("This only works with person Atoms");
                DestroyImmediate(this);
                return;
            }

            // Get Materials from Person
            Atom _person = containingAtom;
            DAZCharacterSelector _selector = _person.GetComponentInChildren<DAZCharacterSelector>();
            materials = _selector.selectedCharacter.skin.GPUmaterials;

            DAZCharacterTextureControl textureControl = _person.GetComponentInChildren<DAZCharacterTextureControl>();

            // Save Reference to Original Base Texture
            originalFaceTexture = materials[faceMaterialSlots[0]].GetTexture("_MainTex");
            secondTexture = originalFaceTexture;


            // Setup Texture Pickers
            RegisterUrl(secondTextureURL);
            RegisterUrl(maskTextureURL);

            UIDynamicButton buttonSecondTex = CreateButton("Browse Overlay Texture", false);
            buttonSecondTex.button.onClick.AddListener(() =>
            {
                // Todo: Most recent DIR, or default
                List<ShortCut> shortCuts = FileManagerSecure.GetShortCutsForDirectory("Custom/Atom/Person/Textures", false, true, true, false);
                SuperController.singleton.GetMediaPathDialog(LoadSecondTexture, "jpg|png|tif|tiff", "Custom/Atom/Person/Textures", false, true, false, null, true, shortCuts, true, true);

            });
            UIDynamicTextField secondTextureLabel = CreateTextField(secondTextureURL, false);

            UIDynamicButton buttonMaskTex = CreateButton("Browse Mask Texture", false);
            buttonMaskTex.button.onClick.AddListener(() =>
            {
                // Todo: Most recent DIR, or default
                string defaultPath = GetPackagePath(this) + DEFAULT_MASK_PATH;
                List<ShortCut> shortCuts = FileManagerSecure.GetShortCutsForDirectory("Custom/Atom/Person/Textures", false, true, true, false);
                SuperController.singleton.GetMediaPathDialog(LoadMaskTexture, "jpg|png|tif|tiff", defaultPath, false, true, false, null, true, shortCuts, true, true);

            });
            UIDynamicTextField maskTextureLabel = CreateTextField(maskTextureURL, false);

            // Setup HSL variables
            RegisterFloat(hue);
            CreateSlider(hue).slider.onValueChanged.AddListener((float val) => { UpdateRenderTexture(); });
            RegisterFloat(saturation);
            CreateSlider(saturation).slider.onValueChanged.AddListener((float val) => { UpdateRenderTexture(); });
            RegisterFloat(value);
            CreateSlider(value).slider.onValueChanged.AddListener((float val) => { UpdateRenderTexture(); });

            UIDynamicButton newTextureButton = CreateButton("Bake Texture", false);
            newTextureButton.button.onClick.AddListener(() =>
            {
                SuperController.singleton.fileBrowserUI.defaultPath = "Custom/Atom/Person/Textures";
                SuperController.singleton.fileBrowserUI.showFiles = true;
                SuperController.singleton.fileBrowserUI.showDirs = true;
                SuperController.singleton.fileBrowserUI.SetTextEntry(true);
                SuperController.singleton.fileBrowserUI.hideExtension = true;
                SuperController.singleton.fileBrowserUI.shortCuts = new List<ShortCut>();
                SuperController.singleton.fileBrowserUI.fileFormat = "png";
                SuperController.singleton.fileBrowserUI.Show(SaveRenderTexture);
            });



            // Load Shader from Asset Bundle
            AssetLoader.AssetBundleFromFileRequest req = new AssetLoader.AssetBundleFromFileRequest()
            {
                path = GetPackagePath(this) + ASSET_BUNDLE_LOCATION,
                callback = LoadAssetBundle,
            };
            AssetLoader.QueueLoadAssetBundleFromFile(req);


        }
        catch (Exception e)
        {
            SuperController.LogError("Exception caught: " + e);
        }
    }



    private void SaveRenderTexture(string path)
    {
        UpdateRenderTexture();



        Texture2D tex = new Texture2D(currentRenderTexture.height, currentRenderTexture.width, TextureFormat.RGB24, false);
        RenderTexture previous = RenderTexture.active;
        RenderTexture.active = currentRenderTexture;
        tex.ReadPixels(new Rect(0, 0, currentRenderTexture.width, currentRenderTexture.height), 0, 0);
        tex.Apply();
        RenderTexture.active = previous;
        FileManagerSecure.WriteAllBytes(path + ".png", tex.EncodeToPNG());
    }

    // Loads Second Texture from URL and then updates Render Texture
    private void LoadSecondTexture(string path)
    {
        secondTextureURL.SetFilePath(path);
        ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
        queuedImage.imgPath = path;
        queuedImage.forceReload = true;
        queuedImage.skipCache = true;
        queuedImage.compress = false;
        queuedImage.createMipMaps = false;
        queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
        {
            secondTexture = qi.tex;
            UpdateRenderTexture();
        };
        ImageLoaderThreaded.singleton.QueueImage(queuedImage);

    }

    private void LoadMaskTexture(string path)
    {
        maskTextureURL.SetFilePath(path);
        ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
        queuedImage.imgPath = path;
        queuedImage.forceReload = true;
        queuedImage.skipCache = true;
        queuedImage.compress = false;
        queuedImage.createMipMaps = false;
        queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
        {
            maskTexture = qi.tex;
            UpdateRenderTexture();
        };
        ImageLoaderThreaded.singleton.QueueImage(queuedImage);
    }

    // Load Shader and face mask from asset bundle
    private void LoadAssetBundle(MeshVR.AssetLoader.AssetBundleFromFileRequest request)
    {
        if (request.assetBundle == null)
        {
            SuperController.LogError("Unable to load Asset Bundle - northernshikima_faceblend.assetbundle");
            return;
        }
        unlitBlendShader = request.assetBundle.LoadAsset<Shader>("assets/shaders/UnlitBlendTextures.shader");


        // If Assets successfully loaded
        if (unlitBlendShader)
        {
            assetsLoaded = true;


            if (maskTextureURL.val == string.Empty)
            {
                maskTextureURL.SetFilePath(GetPackagePath(this) + DEFAULT_MASK_PATH + DEFAULT_MASK_TEXTURE);

            }

            LoadMaskTexture(maskTextureURL.val);
            if (secondTextureURL.val != string.Empty)
            {
                LoadSecondTexture(secondTextureURL.val);

            } else
            {
                UpdateRenderTexture();
            }

        }
    }

    private void UpdateRenderTexture()
    {
        if (!assetsLoaded)
        {
            return;
        }


        RenderTexture newRenderTexture = new RenderTexture(originalFaceTexture.height, originalFaceTexture.width, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);
        newRenderTexture.enableRandomWrite = true;
        newRenderTexture.Create();


        Material blendMaterial = new Material(unlitBlendShader);
        blendMaterial.SetTexture("_BaseTex", originalFaceTexture);
        blendMaterial.SetTexture("_SecondTex", secondTexture);
        blendMaterial.SetTexture("_MaskTex", maskTexture);
        blendMaterial.SetFloat("_Hue", hue.val + 0.5f);
        blendMaterial.SetFloat("_Saturation", saturation.val * 2);
        blendMaterial.SetFloat("_Value", value.val * 2);

        Graphics.Blit(null, newRenderTexture, blendMaterial);

        foreach (int i in faceMaterialSlots)
        {
            materials[i].SetTexture("_MainTex", newRenderTexture);
        }
        // Cleanup old render texture to avoid GPU memory leak
        if (currentRenderTexture)
        {
            currentRenderTexture.Release();
        }
        currentRenderTexture = newRenderTexture;
    }



    public void OnDestroy()
    {
        // Restore old face texture
        foreach (int i in faceMaterialSlots)
        {
            materials[i].SetTexture("_MainTex", originalFaceTexture);
        }
        // Clean up used textures
        Destroy(maskTexture);
        if (secondTexture != originalFaceTexture)
        {
            Destroy(secondTexture);
        }
        // Clean up Render Texture
        currentRenderTexture.Release();

        // Remove lock from bundle
        AssetLoader.DoneWithAssetBundleFromFile(GetPackagePath(this) + ASSET_BUNDLE_LOCATION);
    }

    // Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
    // Copied from MacGruber's Utils
    public static string GetPluginPath(MVRScript self)
    {
        string id = self.name.Substring(0, self.name.IndexOf('_'));
        string filename = self.manager.GetJSON()["plugins"][id].Value;
        return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
    }

    // Get path prefix of the package that contains our plugin.
    // Copied from MacGruber's Utils
    public static string GetPackagePath(MVRScript self)
    {
        string filename = GetPluginPath(self);
        int idx = filename.IndexOf(":/");
        if (idx >= 0)
            return filename.Substring(0, idx + 2);
        else
            return string.Empty;
    }
}